from unittest import mock

import pytest

from pogo_core import error
from pogo_core.util import migrate, sql


@pytest.fixture
def _migration_one(migrations):
    p = migrations / "20240317_01_abcde-initial-migration.sql"

    with p.open("w") as f:
        f.write("""
-- initial migration
-- depends:

-- migrate: apply
CREATE TABLE table_one();

-- migrate: rollback
DROP TABLE table_one;
""")


@pytest.fixture
def _migration_two(migrations, _migration_one):
    p = migrations / "20240317_02_12345-second-migration.sql"

    with p.open("w") as f:
        f.write("""
-- second migration
-- depends: 20240317_01_abcde-initial-migration
-- transaction: false

-- migrate: apply
CREATE TABLE table_two();

-- Auto generated
-- From psql

-- migrate: rollback

-- Auto generated
-- From psql
DROP TABLE table_two;
""")


@pytest.fixture
def _broken_apply(migrations, _migration_two):
    p = migrations / "20240318_01_12345-broken-apply.sql"

    with p.open("w") as f:
        f.write("""
-- broker migration
-- depends: 20240317_02_12345-second-migration

-- migrate: apply
CREATE TABLE table_three;

-- migrate: rollback
DROP TABLE table_two;
""")


@pytest.fixture
def _broken_rollback(migrations, _migration_two):
    p = migrations / "20240318_01_12345-broken-rollback.sql"

    with p.open("w") as f:
        f.write("""
-- broker migration
-- depends: 20240317_02_12345-second-migration

-- migrate: apply
CREATE TABLE table_three();

-- migrate: rollback
DROP TABLE table_four;
""")


class Base:
    async def assert_tables(self, db_session, tables):
        stmt = """
        SELECT schemaname, tablename
        FROM pg_tables
        WHERE  schemaname = 'public' or schemaname = 'pogo'
        ORDER BY schemaname, tablename
        """
        results = await db_session.fetch(stmt)

        assert [f"{r['schemaname']}.{r['tablename']}" for r in results] == tables


@pytest.fixture(autouse=True)
def connect_patch_(db_session, monkeypatch):
    monkeypatch.setattr(sql.asyncpg, "connect", mock.AsyncMock(return_value=db_session))


class TestApply(Base):
    @pytest.mark.usefixtures("migrations")
    async def test_no_migrations_applies_pogo_tables(self, migrations, db_session):
        await migrate.apply(db_session, migrations, schema_name="public")

        await self.assert_tables(db_session, ["public._pogo_migration", "public._pogo_version"])

    @pytest.mark.usefixtures("_migration_two")
    async def test_migrations_applied(self, migrations, db_session):
        await migrate.apply(db_session, migrations, schema_name="public")

        await self.assert_tables(
            db_session,
            ["public._pogo_migration", "public._pogo_version", "public.table_one", "public.table_two"],
        )

    @pytest.mark.usefixtures("_migration_two")
    async def test_already_applied_skips(self, migrations, db_session):
        await migrate.apply(db_session, migrations, schema_name="public")
        await migrate.apply(db_session, migrations, schema_name="public")

        await self.assert_tables(
            db_session,
            ["public._pogo_migration", "public._pogo_version", "public.table_one", "public.table_two"],
        )

    @pytest.mark.usefixtures("_migration_two")
    async def test_apply_multiple_schemas(self, migrations, db_session):
        await migrate.apply(db_session, migrations, schema_name="public")
        db_session = await sql.get_connection("", schema_name="pogo", schema_create=True)
        await migrate.apply(db_session, migrations, schema_name="pogo")

        await self.assert_tables(
            db_session,
            [
                "pogo.table_one",
                "pogo.table_two",
                "public._pogo_migration",
                "public._pogo_version",
                "public.table_one",
                "public.table_two",
            ],
        )

    @pytest.mark.usefixtures("_broken_apply")
    async def test_broken_migration_not_applied(self, migrations, db_session):
        with pytest.raises(error.BadMigrationError) as e:
            await migrate.apply(db_session, migrations, schema_name="public")

        await self.assert_tables(
            db_session,
            ["public._pogo_migration", "public._pogo_version", "public.table_one", "public.table_two"],
        )
        assert str(e.value) == "Failed to apply 20240318_01_12345-broken-apply"


class TestRollback(Base):
    @pytest.mark.usefixtures("migrations")
    async def test_no_migrations_applies_pogo_tables(self, migrations, db_session):
        await migrate.rollback(db_session, migrations, schema_name="public")

        await self.assert_tables(db_session, ["public._pogo_migration", "public._pogo_version"])

    @pytest.mark.usefixtures("_migration_two")
    async def test_latest_removed(self, migrations, db_session):
        await migrate.apply(db_session, migrations, schema_name="public")
        await migrate.rollback(db_session, migrations, schema_name="public", count=1)

        await self.assert_tables(db_session, ["public._pogo_migration", "public._pogo_version", "public.table_one"])

    @pytest.mark.usefixtures("_migration_two")
    async def test_all_removed(self, migrations, db_session):
        await migrate.apply(db_session, migrations, schema_name="public")
        await migrate.rollback(db_session, migrations, schema_name="public")

        await self.assert_tables(db_session, ["public._pogo_migration", "public._pogo_version"])

    @pytest.mark.usefixtures("_migration_two")
    async def test_only_specified_schema_rolledback(self, migrations, db_session):
        await migrate.apply(db_session, migrations, schema_name="public")
        db_session = await sql.get_connection("", schema_name="pogo", schema_create=True)
        await migrate.apply(db_session, migrations, schema_name="pogo")

        await migrate.rollback(db_session, migrations, schema_name="pogo", count=1)

        await self.assert_tables(
            db_session,
            [
                "pogo.table_one",
                "public._pogo_migration",
                "public._pogo_version",
                "public.table_one",
                "public.table_two",
            ],
        )

    @pytest.mark.usefixtures("_broken_rollback")
    async def test_broken_rollback_rollsback(self, migrations, db_session):
        await migrate.apply(db_session, migrations, schema_name="public")
        with pytest.raises(error.BadMigrationError) as e:
            await migrate.rollback(db_session, migrations, schema_name="public", count=1)

        await self.assert_tables(
            db_session,
            [
                "public._pogo_migration",
                "public._pogo_version",
                "public.table_one",
                "public.table_three",
                "public.table_two",
            ],
        )
        assert str(e.value) == "Failed to rollback 20240318_01_12345-broken-rollback"



from __future__ import annotations

import typing as t

from pogo_core.util import migrate, sql

if t.TYPE_CHECKING:
    from pathlib import Path

    import asyncpg


async def apply(
    migrations: Path,
    *,
    db: asyncpg.Connection | None = None,
    database_dsn: str | None = None,
    schema_name: str = "public",
    schema_create: bool = False,
) -> None:
    if db is None and database_dsn is None:
        msg = "One of db or database_dsn are required to apply migrations."
        raise ValueError(msg)

    db_ = (
        db
        if db is not None
        else await sql.get_connection(database_dsn, schema_name=schema_name, schema_create=schema_create)
    )
    try:
        await migrate.apply(db_, migrations, schema_name=schema_name)
    finally:
        if db is None:
            await db_.close()


async def rollback(
    migrations: Path,
    *,
    db: asyncpg.Connection | None = None,
    database_dsn: str | None = None,
    schema_name: str = "public",
) -> None:
    if db is None and database_dsn is None:
        msg = "One of db or database_dsn are required to rollback migrations."
        raise ValueError(msg)

    db_ = db if db is not None else await sql.get_connection(database_dsn, schema_name=schema_name)
    try:
        await migrate.rollback(db_, migrations, schema_name=schema_name)
    finally:
        if db is None:
            await db_.close()



from __future__ import annotations

import typing as t

import asyncpg

from pogo_core.migration import Migration

if t.TYPE_CHECKING:
    from pathlib import Path


async def get_connection(
    connection_string: str,
    *,
    schema_name: str = "public",
    schema_create: bool = False,
) -> asyncpg.Connection:
    db = await asyncpg.connect(connection_string)

    if schema_create:
        await create_schema(db, schema_name=schema_name)

    await db.execute(f"SET search_path TO {schema_name}")
    await ensure_pogo_sync(db)

    return db


async def read_migrations(
    migrations_location: Path,
    db: asyncpg.Connection | None,
    *,
    schema_name: str,
) -> list[Migration]:
    applied_migrations = await get_applied_migrations(db, schema_name=schema_name) if db else set()
    return [
        Migration(path.stem, path, applied_migrations)
        for path in migrations_location.iterdir()
        if path.suffix in {".py", ".sql"}
    ]


async def get_applied_migrations(db: asyncpg.Connection, *, schema_name: str) -> set[str]:
    stmt = """
    SELECT
        migration_id
    FROM public._pogo_migration
    WHERE schema_name = $1
    """
    results = await db.fetch(stmt, schema_name)

    return {r["migration_id"] for r in results}


async def ensure_pogo_sync(db: asyncpg.Connection) -> None:
    stmt = """
    SELECT exists (
        SELECT FROM pg_tables
        WHERE  schemaname = 'public'
        AND    tablename  = '_pogo_version'
    );
    """
    r = await db.fetchrow(stmt)
    if r is not None and not r["exists"]:
        stmt = """
        CREATE TABLE public._pogo_migration (
            migration_hash VARCHAR(64),  -- sha256 hash of the migration id
            migration_id VARCHAR(255),   -- The migration id (ie path basename without extension)
            applied TIMESTAMPTZ,         -- When this id was applied
            PRIMARY KEY (migration_hash)
        );
        """
        await db.execute(stmt)

        stmt = """
        CREATE TABLE public._pogo_version (
            version INT NOT NULL PRIMARY KEY,
            installed TIMESTAMPTZ
        );
        """
        await db.execute(stmt)

        stmt = """
        INSERT INTO public._pogo_version (version, installed) VALUES (0, now());
        """
        await db.execute(stmt)

    stmt = "SELECT version FROM public._pogo_version ORDER BY version DESC LIMIT 1;"
    version = await db.fetchval(stmt)

    if version == 0:
        stmt = """
        ALTER TABLE public._pogo_migration
        ADD COLUMN schema_name VARCHAR(64),      -- Host schema for this set of migrations.
        DROP CONSTRAINT _pogo_migration_pkey;
        """
        await db.execute(stmt)

        stmt = """
        UPDATE public._pogo_migration SET schema_name = 'public';
        """
        await db.execute(stmt)

        stmt = """
        ALTER TABLE public._pogo_migration
        ALTER COLUMN schema_name SET NOT NULL,
        ADD PRIMARY KEY (migration_hash, schema_name);
        """
        await db.execute(stmt)

        stmt = """
        INSERT INTO public._pogo_version (version, installed) VALUES (1, now());
        """
        await db.execute(stmt)


async def migration_applied(
    db: asyncpg.Connection,
    migration_id: str,
    migration_hash: str,
    *,
    schema_name: str,
) -> None:
    stmt = """
    INSERT INTO public._pogo_migration (
        migration_hash,
        migration_id,
        schema_name,
        applied
    ) VALUES (
        $1, $2, $3, now()
    )
    """
    await db.execute(stmt, migration_hash, migration_id, schema_name)


async def migration_unapplied(db: asyncpg.Connection, migration_id: str, *, schema_name: str) -> None:
    stmt = """
    DELETE FROM public._pogo_migration
    WHERE migration_id = $1 AND schema_name = $2
    """
    await db.execute(stmt, migration_id, schema_name)


async def create_schema(db: asyncpg.Connection, *, schema_name: str) -> None:
    stmt = f"""
    CREATE SCHEMA IF NOT EXISTS "{schema_name}";
    """

    await db.execute(stmt)


async def drop_schema(db: asyncpg.Connection, *, schema_name: str) -> None:
    stmt = f"""
    DROP SCHEMA IF EXISTS "{schema_name}";
    """

    await db.execute(stmt)



from unittest import mock

import pytest

from pogo_core.util import testing


async def test_apply_no_db_info(cwd):
    with pytest.raises(ValueError, match="One of db or database_dsn are required to apply migrations"):
        await testing.apply(cwd / "migrations")


async def test_apply(monkeypatch, db_session, cwd):
    monkeypatch.setattr(testing.migrate, "apply", mock.AsyncMock())
    await testing.apply(cwd / "migrations", db=db_session)

    assert testing.migrate.apply.call_args == mock.call(
        db_session,
        cwd / "migrations",
        schema_name="public",
    )


async def test_apply_loads_db(monkeypatch, cwd):
    monkeypatch.setattr(testing.migrate, "apply", mock.AsyncMock())
    mock_session = mock.Mock(close=mock.AsyncMock(return_value=None))
    monkeypatch.setattr(testing.sql, "get_connection", mock.AsyncMock(return_value=mock_session))
    await testing.apply(cwd / "migrations", database_dsn="postgres_dsn")

    assert testing.migrate.apply.call_args == mock.call(
        mock_session,
        cwd / "migrations",
        schema_name="public",
    )
    assert mock_session.close.call_count == 1


async def test_rollback_no_db_info(cwd):
    with pytest.raises(ValueError, match="One of db or database_dsn are required to rollback migrations"):
        await testing.rollback(cwd / "migrations")


async def test_rollback(monkeypatch, db_session, cwd):
    monkeypatch.setattr(testing.migrate, "rollback", mock.AsyncMock())
    await testing.rollback(cwd / "migrations", db=db_session)

    assert testing.migrate.rollback.call_args == mock.call(
        db_session,
        cwd / "migrations",
        schema_name="public",
    )


async def test_rollback_loads_db(monkeypatch, cwd):
    monkeypatch.setattr(testing.migrate, "rollback", mock.AsyncMock())
    mock_session = mock.Mock(close=mock.AsyncMock(return_value=None))
    monkeypatch.setattr(testing.sql, "get_connection", mock.AsyncMock(return_value=mock_session))
    await testing.rollback(cwd / "migrations", database_dsn="postgres_dsn")

    assert testing.migrate.rollback.call_args == mock.call(
        mock_session,
        cwd / "migrations",
        schema_name="public",
    )
    assert mock_session.close.call_count == 1



import os
import pathlib

import asyncpg
import pytest

from pogo_core import types  # noqa: F401 Shut up test coverage on the imports in this module
from pogo_core.migration import Migration
from pogo_core.util import sql


@pytest.fixture(autouse=True)
def cwd(tmp_path):
    orig = pathlib.Path.cwd()

    try:
        os.chdir(str(tmp_path))
        yield tmp_path
    finally:
        os.chdir(orig)


@pytest.fixture
def postgres_dsn():
    return os.environ["POSTGRES_DSN"]


@pytest.fixture(autouse=True)
async def db_session(request, postgres_dsn):
    conn = await asyncpg.connect(postgres_dsn)
    tr = conn.transaction()
    await tr.start()
    if request.node.get_closest_marker("nosync") is None:
        await sql.ensure_pogo_sync(conn)
    try:
        yield conn
    finally:
        await tr.rollback()
        await conn.close()


@pytest.fixture
def migrations(cwd):
    p = cwd / "migrations"
    p.mkdir()

    return p


@pytest.fixture
def migration_file_factory(migrations):
    def factory(mig_id, format_, contents):
        p = migrations / f"{mig_id}.{format_}"
        with p.open("w") as f:
            f.write(contents)

        return p

    return factory


@pytest.fixture(autouse=True)
def _clear_migration_tracking():
    # ClassVar remembers history across tests. Clear it every test.
    try:
        yield
    except:  # noqa: E722, S110
        pass
    finally:
        Migration._Migration__migrations = {}



from unittest import mock

import pytest

from pogo_core.util import sql


async def assert_tables(db_session, tables):
    stmt = """
    SELECT tablename
    FROM pg_tables
    WHERE  schemaname = 'public'
    ORDER BY tablename
    """
    results = await db_session.fetch(stmt)

    assert [r["tablename"] for r in results] == tables


async def assert_schemas(db_session, schemas):
    stmt = """
    SELECT schema_name
    FROM information_schema.schemata
    WHERE  schema_name NOT IN ('pg_toast', 'pg_catalog', 'information_schema')
    ORDER BY schema_name
    """
    results = await db_session.fetch(stmt)

    assert [r["schema_name"] for r in results] == schemas


@pytest.fixture(autouse=True)
def connect_patch_(db_session, monkeypatch):
    monkeypatch.setattr(sql.asyncpg, "connect", mock.AsyncMock(return_value=db_session))


@pytest.mark.nosync
async def test_get_connection_syncs_tables(db_session):
    db = await sql.get_connection(db_session)

    await assert_tables(db_session, ["_pogo_migration", "_pogo_version"])
    assert db == db_session


@pytest.mark.nosync
async def test_get_connection_creates_schema(db_session):
    db = await sql.get_connection(db_session, schema_name="unit", schema_create=True)

    await assert_schemas(db_session, ["public", "unit"])
    assert db == db_session


@pytest.mark.nosync
@pytest.mark.parametrize("schema", [None, "unit"])
async def test_ensure_pogo_sync_creates_tables(db_session, schema):
    if schema:
        db_session = await sql.get_connection("", schema_name=schema, schema_create=True)

    await sql.ensure_pogo_sync(db_session)

    await assert_tables(db_session, ["_pogo_migration", "_pogo_version"])


@pytest.mark.nosync
@pytest.mark.parametrize("schema", [None, "unit"])
async def test_ensure_pogo_sync_handles_existing_tables(db_session, schema):
    if schema:
        db_session = await sql.get_connection("", schema_name=schema, schema_create=True)

    await sql.ensure_pogo_sync(db_session)
    await sql.ensure_pogo_sync(db_session)

    await assert_tables(db_session, ["_pogo_migration", "_pogo_version"])


@pytest.mark.parametrize("schema", [None, "unit"])
async def test_migration_applied(db_session, schema):
    if schema:
        db_session = await sql.get_connection("", schema_name=schema, schema_create=True)

    schema = schema or "public"
    await sql.migration_applied(db_session, "migration_id", "migration_hash", schema_name=schema)

    ids = await sql.get_applied_migrations(db_session, schema_name=schema)
    assert ids == {"migration_id"}


@pytest.mark.parametrize("schema", [None, "unit"])
async def test_migration_unapplied(db_session, schema):
    if schema:
        db_session = await sql.get_connection("", schema_name=schema, schema_create=True)

    schema = schema or "public"

    await sql.migration_applied(db_session, "migration_id", "migration_hash", schema_name=schema)
    await sql.migration_unapplied(db_session, "migration_id", schema_name=schema)

    ids = await sql.get_applied_migrations(db_session, schema_name=schema)
    assert ids == set()


async def test_create_schema(db_session):
    await sql.create_schema(db_session, schema_name="unit")

    await assert_schemas(db_session, ["public", "unit"])


async def test_drop_schema(db_session):
    await db_session.execute("CREATE SCHEMA unit")

    await sql.drop_schema(db_session, schema_name="unit")

    await assert_schemas(db_session, ["public"])



from typing import Protocol


class Logger(Protocol):
    def debug(self, msg, *args, **kwargs): ...  # noqa: ANN201, ANN001, ANN002, ANN003

    def info(self, msg, *args, **kwargs): ...  # noqa: ANN201, ANN001, ANN002, ANN003

    def warning(self, msg, *args, **kwargs): ...  # noqa: ANN201, ANN001, ANN002, ANN003

    def warn(self, msg, *args, **kwargs): ...  # noqa: ANN201, ANN001, ANN002, ANN003

    def error(self, msg, *args, **kwargs): ...  # noqa: ANN201, ANN001, ANN002, ANN003

    def exception(self, msg, *args, exc_info=True, **kwargs): ...  # noqa: ANN201, ANN001, ANN002, ANN003



from __future__ import annotations

import contextlib
import logging
import typing as t

from pogo_core import error
from pogo_core.migration import Migration, topological_sort
from pogo_core.util import sql

if t.TYPE_CHECKING:
    from pathlib import Path

    import asyncpg

    from pogo_core.types import Logger

logger_ = logging.getLogger(__name__)


@contextlib.asynccontextmanager
async def transaction(db: asyncpg.Connection, migration: Migration) -> t.AsyncIterator[None]:
    tr = None
    if migration.use_transaction:
        tr = db.transaction()
        await tr.start()

    try:
        yield
    except Exception:
        if tr:  # pragma: no cover
            await tr.rollback()
        raise
    else:
        if tr:
            await tr.commit()


async def apply(
    db: asyncpg.Connection,
    migrations_dir: Path,
    *,
    schema_name: str,
    logger: Logger | logging.Logger | None = None,
) -> None:
    logger = logger or logger_
    await sql.ensure_pogo_sync(db)
    migrations = await sql.read_migrations(migrations_dir, db, schema_name=schema_name)
    migrations = topological_sort([m.load() for m in migrations])

    for migration in migrations:
        try:
            migration.load()
            if not migration.applied:
                logger.warning("Applying %s", migration.id)
                async with transaction(db, migration):
                    await migration.apply(db)
                    await sql.migration_applied(db, migration.id, migration.hash, schema_name=schema_name)
        except Exception as e:  # noqa: PERF203
            msg = f"Failed to apply {migration.id}"
            raise error.BadMigrationError(msg) from e


async def rollback(
    db: asyncpg.Connection,
    migrations_dir: Path,
    *,
    schema_name: str,
    count: int | None = None,
    logger: Logger | logging.Logger | None = None,
) -> None:
    logger = logger or logger_
    await sql.ensure_pogo_sync(db)
    migrations = await sql.read_migrations(migrations_dir, db, schema_name=schema_name)
    migrations = reversed(list(topological_sort([m.load() for m in migrations])))

    i = 0
    for migration in migrations:
        try:
            migration.load()
            if migration.applied and (count is None or i < count):
                logger.warning("Rolling back %s", migration.id)

                async with transaction(db, migration):
                    await migration.rollback(db)
                    await sql.migration_unapplied(db, migration.id, schema_name=schema_name)
                    i += 1
        except Exception as e:  # noqa: PERF203
            msg = f"Failed to rollback {migration.id}"
            raise error.BadMigrationError(msg) from e
